function [acov1, acov2] = asymp_cov_PWGP(design, regr, tphi, tsigma, ttheta, trndErr)
% this function aims to caluclate the asymptotic distribution of parameter
% estiamte for PWGP
% Return: acov1 and acov2 are the asymptotic covariance for Hypersphere
% decomposition and original crosscorrelation matrix T respectively
% it's corresponding to the order of the parameter (beta, phi, sigma, randomErr, theta/T)

    % get J and H as defined in the equation (7) and (8) of the paper
    tphi = tphi(:)';   tsigma = tsigma(:)';     ttheta = ttheta(:)';
    [J, H] = asymp_cov(design, regr, tphi, tsigma, ttheta, trndErr);
    l = length(design);
    B1 = average_coeff(l, 2*size(H,1)/(l*(l-1)));
    acov1 = B1*(inv(H)*J*inv(H)')*B1';      % parameter of hypersphere decomosition
    
    dT = derv_T(4, ttheta);
    B2 = blkdiag(eye(size(acov1,1)-size(dT,1)),dT');
    acov2 = B2*acov1*B2';   % original parameter of crosscorrelation matrix
end

function [ J, H ] = asymp_cov( design, regr, tphi, tsigma, ttheta, trndErr )
%ASYMP_COV_OPT Summary of this function goes here
%   Detailed explanation goes here
    k = length(design);
    code = 'syms';
    for i = 1 : k
         code = [code ' f' num2str(i) ' s' num2str(i)];
    end
    for i = 1 : k
       for j = 1 : i-1
            code = [code ' a' num2str(i) num2str(j)];
       end
    end
    eval(code);
    syms e d;
    sigma = [];
    alpha = [];
    phi = [];
    code = 'alpha = [';
    for i = 1 : k
       for j = 1 : i-1
            code = [code ' a' num2str(i) num2str(j)];
       end
    end
    code = [code ' ];'];
    eval(code);
    
    code = 'phi = [';
    for i = 1 : k
         code = [code ' f' num2str(i) ','];
    end
    code = [code ' ];'];
    eval(code);
    
    code = 'sigma = [';
    for i = 1 : k
         code = [code ' s' num2str(i) ','];
    end
    code = [code ' ];'];
    eval(code);
    
    
    rndErr = e;
    dis = d;
    para = [phi sigma alpha rndErr];
    value = [tphi tsigma ttheta trndErr];  
    derivs = cell(k,k,length(value));
        
    for i = 1 : k
        design(i).F = feval(regr, design(i).S);
    end
    
    npairs = k*(k-1)/2; 
    np = size(design(1).S,2)*2*(size(phi,1) + size(design(1).F,2) + 1) + 1 + 1;
    ne = size(design(1).S,2)*2*(size(phi,1) + 1) + 1;
    H = zeros(npairs*np);
    J = zeros(npairs*np);
    [T, L] = cross_corr(k, alpha);  
    
    % calculate each block first
    C = uni_cov(design, phi, sigma, T, dis, rndErr, para, value);
    [p, q] = meshgrid(1:k, 1:k);
    pairs = flipud([p(p>=q) q(p>=q)]);
    derivs = cell(size(pairs,1), 1);
    parfor i = 1 : size(pairs,1)
        derivs{i} = diff_uni_cov(design, phi, sigma, alpha, L, dis, rndErr, para, value, pairs(i,1), pairs(i,2));
    end
    dervs = cell(k, k, length(para));
    for i = 1 : size(pairs,1)
        dervs(pairs(i,1), pairs(i,2),:) = derivs{i,:};
        for j = 1 : length(para)
            dervs{pairs(i,2), pairs(i,1),j} = dervs{pairs(i,1), pairs(i,2),j}';
        end
    end
    clear derivs;
    
    % break down into a series
    [p, q] = meshgrid(1:npairs, 1:npairs);
    pairs = flipud([p(p>=q) q(p>=q)]);
    Jps = cell(size(pairs,1), 1);
    Jbs = cell(size(pairs,1), 1);
    parfor i = 1 : size(pairs,1)
        index = pairs(i,:);
        p = index(1);   q = index(2);   
        Cp = cell2mat(C([r(p) s(p)], [r(p) s(p)]));
        Cq = cell2mat(C([r(q) s(q)], [r(q) s(q)]));
        Cpq = cell2mat(C([r(p) s(p)], [r(q) s(q)]));
        inCqp = inv(Cp)*Cpq*inv(Cq);
        
        % para in covariance function
        Jps{i} = asymp_cov_pair(inCqp, phi, sigma, alpha, rndErr, dervs, para, p, q);   
        
        % beta
        Fp = blkdiag(design([r(p) s(p)]).F);
        Fq = blkdiag(design([r(q) s(q)]).F);
        Jbs{i} = Fp'*inCqp*Fq;  
        disp(['--------------' num2str([p,q]) '--------------']);
    end
    for i = 1 : size(pairs,1)
        index = pairs(i,:);
        p = index(1);   q = index(2); 
        Jp = Jps{i};
        Jb = Jbs{i};
        % all parameters
        if p == q
            tH = Jp([1:ne,end],[1:ne,end]);
            H(np*(p-1)+1:np*p,np*(q-1)+1:np*q) = blkdiag(Jb, tH);
            H(np*(p-1)+3:np*p,(1:q-1)*np) = Jp([1:ne, end],ne+(1:q-1));
        end
        J(np*(p-1)+1:np*p, np*(q-1)+1:np*q) = blkdiag(Jb, Jp([1:ne,end],[1:ne,end]));
        J(np*(q-1)+1:np*q, np*(p-1)+1:np*p) = blkdiag(Jb, Jp([1:ne,end],[1:ne,end]))';
    end   
end

function [ aC ] = asymp_cov_pair(inCqp, phi, sigma, alpha, rndErr, dervs, para, p, q)
    Pp = parameter(phi, sigma, rndErr, alpha, p, p==q);
    Pq = parameter(phi, sigma, rndErr, alpha, q, p==q); 
    aC = zeros(length(Pp), length(Pq));
    for i = 1 : length(Pp)
        dCp = diff_sub_cov(dervs, Pp(i), para, p, p);
        for j = 1 : length(Pq)
            dCq = diff_sub_cov(dervs, Pq(j), para, q, q);
            aC(i,j) = 1/2*trace(inCqp'*dCp*inCqp*dCq);
        end
    end
end

function [ dC ] = diff_uni_cov(design, phi, sigma, alpha, L, dis, rndErr, para, value, r, s)  
    for i = 1 : length(para);
        eval([char(para(i)) '=', num2str(value(i)) ';']);
    end
    [S2, S1] = meshgrid(design(s).S, design(r).S); 
    
    dC = repmat({zeros(size(S1))}, size(para));
    if r == s
        p = 0;
    else
        p = s + (r-1)*(r-2)/2;
    end
    Pp = unique([phi([r,s]) sigma([r,s]) rndErr alpha(1:p)]);
    T = L*L';
    for j = 1 : length(Pp)
        dx = Pp(j);
        index = ismember(para, dx);
        dL = diff(L, dx);
        sK = sym_corr(phi, sigma, dis, r, s);
        dsC = diff(sK,dx)* T(r,s) + sK * (dL(r,:)*L(s,:)' + L(r,:)*dL(s,:)');
        sC = vpa(subs(dsC, para, value), 100);
        dC{index} = eval(subs(sC, dis, S2-S1)) + eval(diff(rndErr^2,dx))*eye(size(S2))*(r==s);
    end
end

function [ dC ] = diff_sub_cov(dervs, dx, para, p, q)
    dC = [];
    index = ismember(para, dx);
    for i = [r(p) s(p)]
        dCi = [];
        for j = [r(q) s(q)]
            Cij = dervs{i,j,index};
            dCi = [dCi Cij];
        end
        dC = [dC; dCi];
    end
end

function [ C ] = uni_cov(design, phi, sigma, T, dis, rndErr, para, value)
    for i =1 : length(para);
        eval([char(para(i)) '=', num2str(value(i)) ';']);
    end
    
    k = length(phi);
    C = cell(k, k);
    for i = 1 : k
        for j = 1 : k
            sC = sym_corr(phi, sigma, dis, i, j)*T(i,j);
            sC = vpa(subs(sC, para, value), 100);
            [S2, S1] = meshgrid(design(j).S, design(i).S); 
            Cij = eval(subs(sC, dis, S2-S1)) + eval(rndErr^2)*eye(size(S1))*(i==j);
            C{i,j} = Cij;
        end
    end
end

function [ r ] = sym_corr(phi, sigma, dis, i, j)
    normalizedC = ((phi(i)+phi(j))/2*(1/phi(i)+1/phi(j))/2)^(1/4);
    phiIJ = 2*phi(i)*phi(j) / (phi(i)+phi(j));
    r = exp(-(dis)^2 * phiIJ) / normalizedC * sigma(i)* sigma(j);
end

function [P] = parameter(phi, sigma, rndErr, alpha, p, f)  
    if f
        P = [phi(r(p)) phi(s(p)) sigma(r(p)) sigma(s(p)) rndErr alpha(1:p)];
    else 
        P = [phi(r(p)) phi(s(p)) sigma(r(p)) sigma(s(p)) rndErr alpha(p)];
    end
end

function [T, L] = cross_corr(k, alpha)
    L = sym(zeros(k));
    L(1,1) = sym(1);
    for i = 2 : k
        tL = 1;
        for j = 1 : i-1
            p = (i - 2)*(i - 1)/2 + j;
            L(i,j) = cos(alpha(p))*tL;
            tL = tL*sin(alpha(p));    
        end
        L(i,i) = tL;
    end
    T = L*L';
end

function i = r(p)
    low = ceil(sqrt(2*p+0.25)+0.5);
    up = floor(sqrt(2*p-1.75)+1.5);
    if low == up
        i = low;
    else 
        error('');
    end
end

function j = s(p)
    low = ceil(sqrt(2*p+0.25)+0.5);
    up = floor(sqrt(2*p-1.75)+1.5);
    if low == up
        i = low;
    else 
        error('');
    end
    j = p - (i-1)*(i-2)/2;
end