function [ dmodel, perf ] = dacefit_MV_Given( design, regr, corr, theta0, beta, rndErr)
    
    nInput = 6;
    if nargin < nInput
        error('insufficient number of inputs');
    end
    if  any(theta0 < 0)
        error('theta0 must be strictly positive');
    end        
    if isempty(design)
        error('input design data cannot be empty')
    end
    
    % number of levels
    numLevels = length(design);
    numDataset = size(design(1).Y,2);
    % dimension of input space
    p = size(design(1).S,2);
    
    % check data
    for i = 1 : numLevels
        if isempty(design(i).S)
            error('no design data S at %dth Level. Data cannot be empty', i);
        else
            if size(design(i).S, 2) ~= p
                error('inncorrect dimension of response data S at %dth Level data', i);
            end
        end
        
        if isempty(design(i).Y)
            error('no design data at %dth Level. S cannot be empty', i);
        else
            if size(design(i).Y, 2) ~= numDataset
                error('inncorrect dimension of response data Y at %dth Level data', i);
            end
        end 
    end
    
    
    % Normalization -- remain the same
    for i = 1 : numLevels 
        mS = zeros(1, size(design(i).S,2));
        sS = ones(1, size(design(i).S,2));
        design(i).Ssc = [mS; sS];
        mY = 0;   sY = 1;
        design(i).Ysc = [mY; sY];
    end
        
    design0 = design;
    S = [];     % design sites
    indexLevels = [];
    for i = 1 : numLevels
        S = [S; design(i).S];     
        design(i).F = feval(regr, design(i).S);
        indexLevels = [indexLevels; i*ones(size(design(i).S, 1), 1)];
        design(i).Y = design(i).Y';
    end
    indexLP = indexLevels;   % index of the levels where a point is
    
    % Calculate distances between points
    indexI = (1 : size(S,1))';
    indexJ = (1 : size(S,1))';
    [indexI, indexJ] = meshgrid(indexI, indexJ);
    [indexLevelsI, indexLevelsJ] = meshgrid(indexLevels,indexLevels);
    index = indexI < indexJ;
    indexI = indexI(index);
    indexJ = indexJ(index);
    indexLevelsI = indexLevelsI(index);
    indexLevelsJ = indexLevelsJ(index);
    D = S(indexI,:) - S(indexJ,:);    % distances between points    
    indexIJ = [indexI indexJ];      % index matrix
    % matrix storing index of levels
    indexLev = [indexLevelsI indexLevelsJ];

    par = struct('design',design, 'corr',corr, 'D',D, 'indexIJ',indexIJ, ...
         'indexLev',indexLev, 'indexLP',indexLP, 'beta',beta, 'rndErr',rndErr);

    if nargin == nInput
        % evaluate the object function directly
        [f, fit] = objfunc(theta0, par);
        theta = theta0';
        perf = struct('perf',[theta; f; 1], 'nv',1);
    end
        
    dmodel = struct('regr',regr, 'corr',corr, 'theta',theta(:)', 'C',fit.C, ...
          'beta',fit.beta, 'sigma',fit.sigma, 'gamma',fit.gamma, 'cov',fit.cov, 'phi',fit.phi, ...
          'omega',fit.omega, 'design',design0, 'rndErr',fit.rndErr, 'CC',fit.CC, 'likelihood',f);

% >>>>>>>>>>>>>>>>   Auxiliary functions  ====================

function  [obj, fit] = objfunc(theta, par)
    % Initialize
    obj = inf; 
    
    % total number of design sites
    n = length(par.indexLP);
    k = length(par.design);
    numDataset = size(par.design(1).Y',2);
    
    % set parameters
    sigma = theta(end-k+1:end)';
    theta = theta(1:end-k);
    % set up  R
    [r, CC] = feval(par.corr,  theta, par.D, par.indexLev);
    r = r/par.rndErr^2.*sigma(par.indexLev(:,1)).*sigma(par.indexLev(:,2));
    
    idx = find(r);   o = (1 : n)';   
    R = sparse([par.indexIJ(idx,1); o], [par.indexIJ(idx,2); o], ...
        [r(idx); sigma(par.indexLP).^2/par.rndErr^2+ones(n,1)]);
     
    % Cholesky factorization with check for pos. def.
    [C, rd] = chol(R);
    if  rd,  return, end % not positive definite
    
    C = C';  
    Ft = C \ blkdiag(par.design(:).F);
    invF = Ft'*Ft;
    % check F
    if cond(invF) > 1e30
        warning('F is too ill conditioned\nPoor combination of regression model and design sites');
    end
    
    Yt = C \ [par.design(:).Y]';
    meanYt = mean(Yt,2);
    beta = par.beta;
    res = Yt - repmat(Ft*beta, 1, numDataset);
    rndErr2 = trace(res'*res)/numDataset/n;

    obj = n*log(rndErr2) + 2*sum(log(full(diag(C))));  
    obj = numDataset*obj;
    
    if  nargout > 1
        phi = theta(1:k*size(par.D,2))';
        omega = theta(k*size(par.D,2)+1:end)';
        R = full(R + R'-diag(diag(R)));
        gamma = (Yt - repmat(Ft*beta,1,numDataset));
        gamma = gamma' / C;
        fit = struct('sigma',sigma, 'beta',beta, 'gamma', gamma', 'phi',phi, ...
            'omega',omega, 'C',C, 'cov',rndErr2*R, 'CC',CC, 'rndErr',par.rndErr);
    end
