function [ dmodel, perf ] = dacefit_MV_Sep( design, regr, corr, theta0, lob, upb, varargin)
% this function is to fit the MGP model using separable covariance function
% design: is a strcut has two fileds, one is S, wihch is the inputs for
% MGP, and another is Y, the responses for S
% regr: is the handle of regresion functions
% corr: is the handle of covariance functions
% theta0: is the initial values for searching algorithms
% lob: is the lower bound for searching algorithms
% upb: is the upper bound for searching algorithms
% varargin: is the mothod to specify the search algorithms (dace and fmincon, etc.)    
    nInput = 4;
    if nargin < nInput
        error('insufficient number of inputs');
    end
    if nargin == nInput
        if  any(theta0 < 0)
            error('theta0 must be strictly positive');
        end
        lob = [];
        upb = [];
    else
        % check dimensions
        lth = length(theta0);
        if  length(lob) ~= lth || length(upb) ~= lth
            error('theta0, lob and upb must have the same length');
        end
        if  any(lob < 0) || any(upb < lob)
            error('The bounds must satisfy  0 < lob <= upb')
        end

        % specify search algorithm
        if isempty(varargin)
            SA = 0;
            disp('Using the default search algorithm "dace"!');
        else
            SA = 1;
            srchAlg = varargin{1};
            options = optimset('Display','off');
            if length(varargin) == 2
                options = varargin{2}; 
            end
        end
    end
        
    if isempty(design)
        error('input design data cannot be empty')
    end
    
    % number of levels
    numLevels = length(design);
    numDataset = size(design(1).Y,2);
    % dimension of input space
    p = size(design(1).S,2);
    
    % check data
    for i = 1 : numLevels
        if isempty(design(i).S)
            error('no design data S at %dth Level. Data cannot be empty', i);
        else
            if size(design(i).S, 2) ~= p
                error('inncorrect dimension of response data S at %dth Level data', i);
            end
        end
        
        if isempty(design(i).Y)
            error('no design data at %dth Level. S cannot be empty', i);
        else
            if size(design(i).Y, 2) ~= numDataset
                error('inncorrect dimension of response data Y at %dth Level data', i);
            end
        end 
    end
    
    
    % Normalization -- remain the same
    for i = 1 : numLevels 
        mS = zeros(1, size(design(i).S,2));
        sS = ones(1, size(design(i).S,2));
        design(i).Ssc = [mS; sS];
        mY = 0;   sY = 1;
        design(i).Ysc = [mY; sY];
    end
        
    design0 = design;
    S = [];     % design sites
    indexLevels = [];
    for i = 1 : numLevels
        S = [S; design(i).S];     
        design(i).F = feval(regr, design(i).S);
        indexLevels = [indexLevels; i*ones(size(design(i).S, 1), 1)];
        design(i).Y = design(i).Y';
    end
    indexLP = indexLevels;   % index of the levels where a point is
    
    % Calculate distances between points
    indexI = (1 : size(S,1))';
    indexJ = (1 : size(S,1))';
    [indexI, indexJ] = meshgrid(indexI, indexJ);
    [indexLevelsI, indexLevelsJ] = meshgrid(indexLevels,indexLevels);
    index = indexI < indexJ;
    indexI = indexI(index);
    indexJ = indexJ(index);
    indexLevelsI = indexLevelsI(index);
    indexLevelsJ = indexLevelsJ(index);
    D = S(indexI,:) - S(indexJ,:);    % distances between points    
    indexIJ = [indexI indexJ];      % index matrix
    % matrix storing index of levels
    indexLev = [indexLevelsI indexLevelsJ];

    par = struct('design',design, 'corr',corr, 'D',D, ...
        'indexIJ',indexIJ, 'indexLev',indexLev, 'indexLP',indexLP);

    if nargin == nInput
        % evaluate the object function directly
        [f, fit] = objfunc(theta0, par);
        theta = theta0';
        perf = struct('perf',[theta; f; 1], 'nv',1);
    else
        if SA == 0
            % pattern search method originally provided in DACE package
            [theta, f, fit, perf] = boxmin(theta0, lob, upb, par);
        else
             % search algorithm in matlab e.g. 'fmincon' and 'fminbnd'
            [theta, f, fit, perf] = mleparam(theta0, lob, upb, par, srchAlg, options);
        end
        if  isinf(f)
            error('Bad parameter region.  Try adjuest the boundaries');
        end
    end
        
    dmodel = struct('regr',regr, 'corr',corr, 'theta',theta(:)', 'C',fit.C, ...
          'beta',fit.beta, 'sigma',fit.sigma, 'gamma',fit.gamma, 'cov',fit.cov,  'phi',fit.phi, ...
          'omega',fit.omega, 'design',design0, 'rndErr',fit.rndErr, 'CC',fit.CC, 'likelihood',f);

% >>>>>>>>>>>>>>>>   Auxiliary functions  ====================
function  [obj, fit] = objfunc(theta, par)
    % Initialize
    obj = inf; 
    
    % total number of design sites
    n = length(par.indexLP);
    k = length(par.design);
    numDataset = size(par.design(1).Y',2);
    
    % set parameters
    sigma = theta(end-k+1:end)';
    phi = theta(1:size(par.D,2));
    omega = theta(size(par.D,2)+1:end-k);
    theta = [repmat(phi,1,k) omega];
    % set up  R
    [r, CC] = feval(par.corr, theta, par.D, par.indexLev);
    r = r.*sigma(par.indexLev(:,1)).*sigma(par.indexLev(:,2));
    
    idx = find(r);   o = (1 : n)';   
    R = sparse([par.indexIJ(idx,1); o], [par.indexIJ(idx,2); o], ...
        [r(idx); sigma(par.indexLP).^2+ones(n,1)]); 
    
    % Cholesky factorization with check for pos. def.
    [C, rd] = chol(R);
    if  rd,  return, end % not positive definite
    
    C = C';  
    Ft = C \ blkdiag(par.design(:).F);
    invF = Ft'*Ft;
    % check F
    if cond(invF) > 1e30
        warning('F is too ill conditioned\nPoor combination of regression model and design sites');
    end
    
    Yt = C \ [par.design(:).Y]';
    meanYt = mean(Yt,2);
    beta = invF \ Ft'*meanYt;
    res = Yt - repmat(Ft*beta, 1, numDataset);
    rndErr2 = trace(res'*res)/numDataset/n;

    obj = n*log(rndErr2) + 2*sum(log(full(diag(C))));  
    obj = numDataset*obj;
    
    if  nargout > 1
        R = full(R + R'-diag(diag(R)));
        gamma = (Yt - repmat(Ft*beta,1,numDataset));
        gamma = gamma' / C;
        sigma = sigma*sqrt(rndErr2);
        fit = struct('sigma',sigma, 'beta',beta, 'gamma', gamma', 'phi',phi, ...
            'omega',omega, 'C',C, 'cov',rndErr2*R, 'CC',CC, 'rndErr',sqrt(rndErr2));
    end

% --------------------------------------------------------

function  [t, f, fit, perf] = boxmin(t0, lo, up, par)
%BOXMIN  Minimize with positive box constraints

% Initialize
[t, f, fit, itpar] = start(t0, lo, up, par);
if  ~isinf(f)
  % Iterate
  p = length(t);
  if  p <= 2,  kmax = 2; else kmax = 4*min(p,4); end
  for  k = 1 : kmax
    th = t;
    [t, f, fit, itpar] = explore(t, f, fit, itpar, par);
    [t, f, fit, itpar] = move(th, t, f, fit, itpar, par);
  end
end
perf = struct('nv',itpar.nv, 'perf',itpar.perf(:,1:itpar.nv));

% --------------------------------------------------------

function  [t, f, fit, itpar] = start(t0, lo, up, par)
% Get starting point and iteration parameters

% Initialize
t = t0(:);  lo = lo(:);   up = up(:);   p = length(t);
D = 2 .^ ((1:p)'/(p+2));
ee = find(up == lo);  % Equality constraints
if  ~isempty(ee)
  D(ee) = ones(length(ee),1);   t(ee) = up(ee); 
end
ng = find(t < lo | up < t);  % Free starting values
if  ~isempty(ng)
  t(ng) = (lo(ng) .* up(ng).^7).^(1/8);  % Starting point
end
ne = find(D ~= 1);

% Check starting point and initialize performance info
[f, fit] = objfunc(t,par);   nv = 1;
itpar = struct('D',D, 'ne',ne, 'lo',lo, 'up',up, ...
  'perf',zeros(p+2,200*p), 'nv',1);
itpar.perf(:,1) = [t; f; 1];
if  isinf(f)    % Bad parameter region
  return
end

if  length(ng) > 1  % Try to improve starting guess
  d0 = 16;  d1 = 2;   q = length(ng);
  th = t;   fh = f;   jdom = ng(1);  
  for  k = 1 : q
    j = ng(k);    fk = fh;  tk = th;
    DD = ones(p,1);  DD(ng) = repmat(1/d1,q,1);  DD(j) = 1/d0;
    alpha = min(log(lo(ng) ./ th(ng)) ./ log(DD(ng))) / 5;
    v = DD .^ alpha;   tk = th;
    for  rept = 1 : 4
      tt = tk .* v; 
      [ff, fitt] = objfunc(tt,par);  nv = nv+1;
      itpar.perf(:,nv) = [tt; ff; 1];
      if  ff <= fk 
        tk = tt;  fk = ff;
        if  ff <= f
          t = tt;  f = ff;  fit = fitt; jdom = j;
        end
      else
        itpar.perf(end,nv) = -1;   break
      end
    end
  end % improve
  
  % Update Delta  
  if  jdom > 1
    D([1 jdom]) = D([jdom 1]); 
    itpar.D = D;
  end
end % free variables

itpar.nv = nv;

% --------------------------------------------------------

function  [t, f, fit, itpar] = explore(t, f, fit, itpar, par)
% Explore step

nv = itpar.nv;   ne = itpar.ne;
for  k = 1 : length(ne)
  j = ne(k);   tt = t;   DD = itpar.D(j);
  if  t(j) == itpar.up(j)
    atbd = 1;   tt(j) = t(j) / sqrt(DD);
  elseif  t(j) == itpar.lo(j)
    atbd = 1;  tt(j) = t(j) * sqrt(DD);
  else
    atbd = 0;  tt(j) = min(itpar.up(j), t(j)*DD);
  end
  [ff, fitt] = objfunc(tt,par);  nv = nv+1;
  itpar.perf(:,nv) = [tt; ff; 2];
  if  ff < f
    t = tt;  f = ff;  fit = fitt;
  else
    itpar.perf(end,nv) = -2;
    if  ~atbd  % try decrease
      tt(j) = max(itpar.lo(j), t(j)/DD);
      [ff, fitt] = objfunc(tt,par);  nv = nv+1;
      itpar.perf(:,nv) = [tt; ff; 2];
      if  ff < f
        t = tt;  f = ff;  fit = fitt;
      else
        itpar.perf(end,nv) = -2;
      end
    end
  end
end % k

itpar.nv = nv;

% --------------------------------------------------------

function  [t, f, fit, itpar] = move(th, t, f, fit, itpar, par)
% Pattern move

nv = itpar.nv;   ne = itpar.ne;   p = length(t);
v = t ./ th;
if  all(v == 1)
  itpar.D = itpar.D([2:p 1]).^.2;
  return
end

% Proper move
rept = 1;
while  rept
  tt = min(itpar.up, max(itpar.lo, t .* v));  
  [ff, fitt] = objfunc(tt,par);  nv = nv+1;
  itpar.perf(:,nv) = [tt; ff; 3];
  if  ff < f
    t = tt;  f = ff;  fit = fitt;
    v = v .^ 2;
  else
    itpar.perf(end,nv) = -3;
    rept = 0;
  end
  if  any(tt == itpar.lo | tt == itpar.up), rept = 0; end
end

itpar.nv = nv;
itpar.D = itpar.D([2:p 1]).^.25;

function [theta, fmin, fit, output] = mleparam(theta0, lb, ub, par, srchAlg, options)
% Matlab Optimization Toolbox is needed
    ofun = @(theta)objfunc(theta,par);
    switch srchAlg
        case 'fmincon'
           [theta, fmin, ~, output] = fmincon(ofun,theta0,[],[],[],[],lb,ub,[],options);
        case 'fminsearch'
           [theta, fmin, ~, output] = fminsearchbnd(ofun, theta0, lb,ub, options);
        otherwise
            error('wrong input parameters for search algorithms');
    end      
    [~, fit] = objfunc(theta, par);