function [ dmodel, perf ] = dacefit_Pairwise( design, regr, corr, theta0, LB, UB, varargin )
% this function is to fit the PWGP model using nonseparable covariance
% function by optimizing composite likelhood
% design: is a strcut has two fileds, one is S, wihch is the inputs for
% PWGP, and another is Y, the responses for S
% regr: is the handle of regresion functions
% corr: is the handle of covariance functions
% theta0: is the initial values for searching algorithms
% LB: is the lower bound for searching algorithms
% UP: is the upper bound for searching algorithms
% varargin: is the mothod to specify the search algorithms (dace and fmincon, etc.)
    nInput = 6;
    if nargin < nInput
        error('insufficient number of inputs');
    end
    if nargin == nInput
        if  any(theta0 < 0)
            error('theta0 must be strictly positive');
        end
    else
        % check dimensions
        lth = length(theta0);
        if  length(LB) ~= lth || length(UB) ~= lth
            error('theta0, lob and upb must have the same length');
        end
        if  any(LB < 0) || any(UB < LB)
            error('The bounds must satisfy  0 < lob <= upb')
        end 
        
        % specify search algorithm
        if isempty(varargin)
            SA = 0;
        else
            SA = 1;
            srchAlg = varargin{1};
            options = optimset('Display','off');
            if length(varargin) == 2
                options = varargin{2}; 
            end
        end
    end
        
    if isempty(design)
        error('input design data cannot be empty')
    end
    
    % init setting
    l = length(design);
    p = size(design(1).S,2);
    q = length(regr(zeros(1,p)));
    initPhi = theta0(1:l*p);
    initTheta = theta0(1+l*p:l*(2*p+l-1)/2);
    initSigma = theta0(l*(2*p+l-1)/2+1:end);
    
    % pairwise method
    count = zeros(l, 1);
    pwSigma = zeros(l, l-1);
    pwBeta = zeros(l*q, l-1);
    pwPhi = zeros(l*p, l-1);
    pwTheta = pi/2 * ones(1, l*(l-1)/2);
    pwRndErr = zeros(1, l*(l-1)/2);
    for i = 1 : l
        for j = 1 : i-1
            count(i) = count(i) + 1;
            count(j) = count(j) + 1;
            theta0 = [ initPhi([i*p-p+1:i*p j*p-p+1:j*p])    initTheta((i-1)*(i-2)/2 + j)  initSigma([i j])];
            ub = [ UB([i*p-p+1:i*p j*p-p+1:j*p])   UB(l*p+(i-1)*(i-2)/2 + j)   UB([i j]+l*(2*p+l-1)/2)];
            lb = [ LB([i*p-p+1:i*p j*p-p+1:j*p])   LB(l*p+(i-1)*(i-2)/2 + j)   LB([i j]+l*(2*p+l-1)/2)];
            if SA
                [dmodelPW, perfPW] = dacefit_PW(design([i j]), regr, corr{1}, pwTheta, i, j, theta0, lb, ub, srchAlg, options);
            else
                [dmodelPW, perfPW] = dacefit_PW(design([i j]), regr, corr{1}, pwTheta, i, j, theta0, lb, ub);
            end
            pwPhi(i*p-p+1:i*p, count(i)) = dmodelPW.phi(1:p)';
            pwPhi(j*p-p+1:j*p, count(j)) = dmodelPW.phi(p+1:2*p)';
            pwSigma(i, count(i)) = dmodelPW.sigma(1);
            pwSigma(j, count(j)) = dmodelPW.sigma(2);
            pwBeta(i*q-q+1:i*q, count(i)) = dmodelPW.beta(1);
            pwBeta(j*q-q+1:j*q, count(j)) = dmodelPW.beta(2);
            pwTheta((i-1)*(i-2)/2 + j) = dmodelPW.omega;
            pwRndErr((i-1)*(i-2)/2 + j) = dmodelPW.rndErr;
        end
    end
    theta0 = [mean(pwPhi') pwTheta mean(pwSigma')];
    [dmodelPW1, perfPW1] = dacefit_MV_Given(design, regr, corr{2}, theta0, mean(pwBeta,2), mean(pwRndErr)); 
    dmodelPW1.pwRndErr = pwRndErr;
    dmodel = dmodelPW1;
    perf = perfPW1;

end

function [ub, lb] = get_bound(T, Li, Lj, ub, lb)
    % matrix construction based on hypersphere decomposition
    % T:  parameters 
    
    % maxbound
    b = 0.99;
    if isempty(T)
        L = 1;
        return;
    end
    
    J = ceil(sqrt(2*length(T)));  % number of levels
    % if J*(J-1)/2 ~= length(T)  error('Incorrect number of parameters!'), end
    L = zeros(J,J); % initialize L matrix

    L(1,1) = 1;
    for i = 2 : Li  % row index in L
        id = (i-1)*(i-2)/2 + 1; % starting parameter for row i
        L(i,1) = 1;
        for j = 2 : i
            L(i,j) = L(i,j-1)*sin(T(id+j-2));
        end
        L(i,1:i-1) = (T(id:id+i-2) ~= pi/2) .* cos(T(id:id+i-2)) .* L(i,1:i-1);
    end
    L(i,1) = 1;
    for j = 2 : Lj
        L(i,j) = L(i,j-1)*sin(T(id+j-2));
    end
    L(i,1:Lj-1) = (T(id:id+Lj-2)~=pi/2) .* cos(T(id:id+Lj-2)) .* L(i,1:Lj-1);

    id = (i-1)*(i-2)/2 + Lj;
    r = L(Lj,:)*L(Li,:)';
    r0 = L(Lj,Lj)*L(Li,Lj);
    if r - r0 + cos(lb(id))*r0 > b && abs((b-r+r0)/r0) < 1
        lb(id) = acos((b-r+r0)/r0);
    end
    if r - r0 + cos(ub(id))*r0 < -b && abs((-b-r+r0)/r0) < 1
        ub(id) = acos((-b-r+r0)/r0);
    end
    
end

